(ns document-storage.core
  (:require [clj-jgit.porcelain :as jp]
            [clojure.tools.logging :as log]
            [clojure.data :refer [diff]]
            [clojure.java.io :as io]
            [clojure.edn :refer [read-string]]
            [clojure.pprint :refer [pprint]]))

(defonce repo-dir (atom ""))
(defonce repositories (atom {}))
(defonce default-repository :default)
(defonce upstream-ssh-identity "upstream")
(defonce storage-directory (atom ""))


(defn init-repository [dir]
  (when-not (jp/discover-repo dir)
    (log/info dir " is not a Git-Repository")
    (log/info "Initialize Git-Repository in " dir)
    (jp/git-init dir))
  (jp/load-repo (jp/discover-repo dir)))

(defn load-config [dir]
  (-> dir
      (str "/config.edn")
      slurp
      read-string))

(defn init-storage [dir]
  (log/info "Init Storage in " dir)
  (reset! storage-directory dir)
  (if-let [config (load-config dir)]
    (doseq [repo (keys (:repositories config))]
      (swap! repositories
             assoc repo
             (merge {:directory (str dir "/" (name repo))
                     :unsaved-files []
                     :git-repo (init-repository (str dir "/" (name repo)))}
                    (get (:repositories config) repo))))
    (log/error "Unable to initialize Document Storage - no config!")))

(defn get-file-path [id repository]
  (let [repo-directory (get-in @repositories [repository :directory])
        file-path (str repo-directory "/" id)]
    file-path))



(defn get-repository-path [repository]
  (let [repo-directory (get-in @repositories [repository :directory])]
    repo-directory))

(defn valid-id? [id repository]
  (let [file-path (get-file-path id repository)]
    (.exists (io/as-file file-path))))

(defn get-repo-data [id]
  (if-let [repo (get-in @repositories [id])]
    repo
    (log/error "invalid repository " id)
    ))

(defn- get-repo
  ([]
   (get-repo default-repository))
  ([id]
   (:git-repo (get-repo-data id))))

(defn- commit-document-change [message repository]
  (jp/git-commit (get-repo repository) message {:name "Document-Manager" :email "burke@fh-muenster.de"}))

(defn save-document [id document & {:keys [repository] :or {repository default-repository}}]
  (let [file-path (get-file-path id repository)
        document (with-out-str (pprint document))]
    (when-not (and (valid-id? id repository) (= (slurp file-path) document))
      (spit file-path document)
      (jp/git-add (get-repo repository) id)
      (commit-document-change (str "saved new version of " id) repository))))

(defn delete-document [id & {:keys [repository] :or {repository default-repository}}]
  (when (valid-id? id repository)
    (io/delete-file (get-file-path id :units) true)
    (jp/git-rm (get-repo repository) id)
    (commit-document-change (str "deleted document " id) repository)))

(defn load-document [id & {:keys [repository] :or {repository default-repository}}]
  (when (valid-id? id repository)
    (read-string (slurp (get-file-path id repository)))))

(defn list-documents [& {:keys [repository] :or {repository default-repository}}]
  (let [repo-path (get-repository-path repository)
        files     (->> (io/file repo-path)
                       .listFiles
                       (filter #(.isFile %))
                       (map #(.getName %))
                      )]
    files))

(defn query-fits? [data query]
  (reduce (fn [a b]
            (and (true? a) (= a b)))

          (map
           (fn [[key [compare-fn val]]]
             (compare-fn (key data) val))
           query)))

(defn find-documents [query & {:keys [repository] :or {repository default-repository}}]
  (let [docs (list-documents :repository repository)]
    (into {} (keep (fn [fname]
                     (let [doc (load-document fname :repository repository)]
                       (when (query-fits? doc query)
                         [fname doc]
                         )
                       ))
                   docs))))

(defn get-upstreams [repo-name]
  (-> (get-repo-data repo-name)
      :upstreams))

(defn init-upstreams [repo-name]
  (let [git (get-repo repo-name)
        upstreams (get-upstreams :default)]
    (doseq [[id {:keys [url]}] upstreams]
      (jp/git-remote-remove git (name id))
      (jp/git-remote-add git (name id) url)
      )
    )
  )


(defn push-upstreams [repo-name]
  (let [git (get-repo repo-name)
        upstreams (jp/git-remote-list git)]
    (jp/with-identity {:name (str @storage-directory "/" upstream-ssh-identity) :exclusive true}
      (doseq [[id urls] upstreams]
        (jp/git-push git :remote id)
        ))
    )
  )

(defn shutdown []
  (pmap (fn [[id repo]]
          (push-upstreams id)
          ) @repositories))

